/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdarg.h>
#include <string.h>

#include "server.h"
#include "crypto.h"

char caller[128];
char sname[128];

char rootdir[MAXPATHLEN];
static char buf[MAXPATHLEN*2 + 300];

int logtostderr = 0;

void
log(char *msg, ...) {
	va_list args;

	va_start(args, msg);
	if (logtostderr) {
		if (logtostderr == 1) {
			fseek(stderr, SEEK_END, 0L);
			logtostderr = 2;
		}
		vfprintf(stderr, msg, args);
		fprintf(stderr, "\n");
	} else {
		char buf[MAXIO];
		vsprintf(buf, msg, args);
		syslog(LOG_INFO, buf);
	}
	va_end(args);
}

int
debugbind(int port) {
	struct	sockaddr_in mya;
	int	x;
	int	nread;
	int	on = 1;

	if((x = socket(AF_INET,SOCK_STREAM,0)) < 0) {
		perror("socket");
		exit(20);
	}
	mya.sin_family = AF_INET;
	memset(&mya.sin_addr, 0, sizeof(mya.sin_addr));
	mya.sin_port = htons(port);
	if(bind(x,(struct sockaddr *)&mya,sizeof(mya))) {
		perror("bind");
		exit(21);
	}
	fprintf(stderr, "listening on port %d\n", port);
	if(listen(x,1) < 0) {
		perror("listen");
		exit(22);
	}
	if((nread = accept(x,0,0)) < 0) {
		perror("accept");
		exit(23);
	}
	if (setsockopt(nread, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof(on)) < 0)
		perror("bind setsockopt");
	if (debug > 1)
		fprintf(stderr, "Connection accepted...\n");
	return nread;
}

char *
getcaller(int fd) {
	struct sockaddr_in p_addr;
	int x;

	/* find out who is calling us */
	x = sizeof(p_addr);
	if(getpeername(fd, (struct sockaddr *)&p_addr, &x) < 0) {
		if (errno == ENOTSOCK)
			return 0;
		else {
			/*
			 * This failure is non-fatal, though some sites
			 * might want it to be.
			 */
			return "<getpeername failed>";
		}
	}
	strncpy(caller, inet_ntoa(p_addr.sin_addr), sizeof(caller) - 2);
	return caller;
}

/*
 * Lookup and setup service-related information.  If we return
 * an error message, it will be reported to the client.
 */
char *
setup_service(char *servicename) {
	int srv_uid = -1;
	int srv_gid = -1;
	char *cp, *ep;

	cp = findservice(conffn, servicename);
	if (cp)
		return cp;

	if ((cp=strtok(0, " \t\n")) == 0)
		return "service missing directory spec";
	if (strlen(cp) >= sizeof(rootdir))
		return "service directory name is too long";
	strcpy(rootdir, cp);

	/*
	 * Confine ourselves...
	 *
	 * If we are running 'unsafe', we can't or won't do the chroot.
	 * We will have to prepend 'rootdir' to all further local file accesses,
	 * and make sure that directory ".." is never used.
	 */
	if (chdir(rootdir)) {
		sprintf(buf,  "chdir %s: %s", rootdir, strerror(errno));
		log(buf);
		return buf;
	}
	if (!unsafe) {
		if (chroot(rootdir)) {
			sprintf(buf,  "chroot %s: %s", rootdir, strerror(errno));
			log(buf);
			return buf;
		}
		rootdir[0] = '/';
	} else
		strcat(rootdir, "/");

	if ((cp=strtok(0, " \t\n")) == 0)
		return 0;
	if (!isdigit(*cp))
		return "service uid is non-numeric";
	srv_uid = strtol(cp, &cp, 10);

	if (*cp == ',') { /* get a gid, too. */
		if (!isdigit(cp[1]))
			return "service gid is non-numeric";
		srv_gid = strtol(cp+1, 0, 10);
	}
	/*
	 * ... and relinquish privilege.
	 */
	if (srv_gid >= 0 && setgid(srv_gid) < 0) {
		sprintf(buf,  "setgid %d: %s", srv_gid, strerror(errno));
		log(buf);
		return buf;
	}
	if (srv_uid >= 0 && setuid(srv_uid) < 0) {
		sprintf(buf,  "setuid %d: %s", srv_uid, strerror(errno));
		log(buf);
		return buf;
	}
	if ((cp=strtok(0, " \t\n")) == 0)
		return 0;
	if (*cp == 'r') {
		srv_readonly = 1;
		return 0;
	}
	if (!isdigit(*cp))
		return "unknown umask";
	srv_umask = strtol(cp, &cp, 0);
	srv_or_perm = 0;
	if (*cp == 0)
		return 0;
	if (*cp++ != ',' || !isdigit(*cp))
		return "bad umask/ormask format";
	srv_or_perm = strtol(cp, &cp, 0);
	if (*cp != 0)
		return "bad umask/ormask termination";
	return 0;
}

/*
 * Determine the full local path of the requested path.  This is
 * easy if we have chrooted to the target directory (unsafe == 0)
 * but harder if we have to prepend "rootdir" and make sure there
 * is no monkey business in the supplied path.  Return NULL if 
 * there is an error.
 */
char *
server_path(char *path, char *localpath) {
	char *cp;

	if (strlen(path)+1 > MAXPATHLEN) {
		sprintf(buf, "file name too long: %s", path);
		return buf;
	}
	if (!unsafe) {
		strcpy(localpath, path);
		return 0;
	}

	strcpy(localpath, rootdir);
	if (strlen(rootdir) + strlen(path) + 2 > MAXPATHLEN) {
		sprintf(buf, "path name too long: %s%s",
			rootdir, path);
		return buf;
	}
	strcpy(localpath, rootdir);
	strcat(localpath, path);
	cp = strstr(localpath, "/..");
	if (cp == 0 || (cp[3] != '\0' && cp[3] != '/'))
		return 0;

	sprintf(buf, "path may contain parent directory: %s", localpath);
	return buf;
}

/*
 * Show the path without our local root.
 */
char *
unroot(char *path) {
	return &path[strlen(rootdir)];
}

/*
 * This is called by crypto_server with the service name desired
 * by the client.  It calls setup_service, which differs for each server.
 */
char *
setservice(char *servicename, KEY key) {
	int n = strlen(servicename);
	char fn[MAXIO];
	char *ep;

	if (debug > 1)
		log("service name: %s", servicename);
	if (n + strlen(keydir) + 2 > MAXIO)
		return "service name/path too long";

	strncpy(sname, servicename, sizeof(sname)-1);
	sname[sizeof(sname)-1] = '\0';
	/*
	 * We use the service name to generate a file name.  Our caller
	 * ensured that there were no slashes in the service name, so
	 * the caller can't wander around our file system.
	 */
	sprintf(fn, "%s/%s", keydir, servicename);
	if ((ep=readkey(fn, key)) != 0)
		return ep;

	return setup_service(servicename);
}
